#!/usr/bin/env python




class atom_info(object):
    '''
    Class for all information attributed to atoms and molecules in the system.
    '''
    def __init__(self, atom_num, atom_name, mass, charge, resid, 
            resname, bonded_to_atom_num, dihedral_list): #order important

        self.atom_num = atom_num
        self.atom_name = atom_name
        self.resid = resid
        self.resname = resname
        self.mass = mass
        self.charge = charge
        self.bonded_to_atom_num = bonded_to_atom_num
        self.dihedral_list = dihedral_list 
            #list of lists of atom nums in dihedrals, not including Hs

        self.molecule_atomNums = None #list of bonded UA atom nums
        self.bonded_to_resids = None #list of resids residue is bonded to

        self.coords = None
        self.forces = None
        self.nearest_sorted_array = None #list of (atom_nums, distances) 
                #from nearest to furthest within a set cutoff distance, 
                #not including Hs
        self.nearest_all_atom_array = None #list containing sorted by 
                #dist (atom_num, dist) for all atoms, including Hs
        self.PKenergy = None
        self.UA_PKenergy = None

        self.bondedUA_H = None #[num_bondedUAs, num_bonded Hs]
        self.nearest_atom = None #HB acceptor all atom info
        self.nearest_Hs = [] #HB
        self.dist = None #HB distance between donor and acceptor
        self.broken = None #HB
        self.RAD = None # list of atoms.all_info
        self.RAD_dist = None # list of (atoms.all_info, dist from ref)

        self.UAweightedForces = None
        self.UAweightedTorques = None
        self.MweightedForces = None
        self.MweightedTorques = None
        self.molecule_UA_Fs = None
        self.molecule_UA_Ts = None
        self.WMprincipalAxis = None

        self.dihedral_phi_type = None #unique dihedral angles, in degrees
        self.nearestAnyNonlike = None #tuple (atom.all_info, dist)
        self.RAD_shell_ranked = None #list of 3 tuple
                #RAD shell ranked, acceptors ranked, donors ranked by RAD
        self.RAD_nAtoms = None
        self.hydrationShellRAD_solute = None #nearest solute molecule centric



    def __repr__(self): #return the output below if whole object is printed
        return "%s %s %s %s %s %s %s" % (self.atom_num, self.atom_name, 
                self.mass, self.charge, self.resid, self.resname, 
                self.bonded_to_atom_num)





def clearClass(all_data):
    '''
    For variables in the class, clear after each frame is analysed.
    '''
    
    def reset(self):
        '''
        reset attributes in a class. Except those from topology.
        This prevents memory overload from each frame.
        '''
        dic = vars(self)
        noEdit = ['atom_num', 'atom_name', 'resid', 'resname', 'mass', 
                'charge', 'bonded_to_atom_num', 'nearest_Hs', 
                'molecule_atomNums', 'bonded_to_resids', 
                'dihedral_list', 'bondedUA_H']
        for i in dic.keys():
            if i not in noEdit:
                dic[i] = None
            else:
                continue
        dic['nearest_Hs'] = []

    # clear all objects from previous frame
    for x in range(0, len(all_data)):
        atom = all_data[x]
        reset(atom)






#!/usr/bin/env python

import sys
import math
import numpy as np
from numpy import linalg as LA
import logging



def distance(x0, x1, dimensions):
    '''
    calculates distance and accounts for PBCs.
    '''

    x0 = np.array(x0)
    x1 = np.array(x1)
    delta = np.abs(x1 - x0)
    delta = np.where(delta > 0.5 * dimensions, delta - dimensions, delta)
    dist = np.sqrt((delta ** 2).sum(axis=-1))
    return dist



def vector(x0, x1, dimensions):
    '''
    get vector of two coordinates over PBCs.
    '''

    x0 = np.array(x0)
    x1 = np.array(x1)
    dimensions = np.array(dimensions)
    delta = x1 - x0
    delta2 = []
    for delt, box in zip(delta, dimensions):
        if delt < 0 and delt < 0.5 * box:
            delt = delt + box
        if delt > 0 and delt > 0.5 * box:
            delt = delt - box

        delta2.append(delt)

    delta2 = np.array(delta2)

    return delta2


def getMcoords(x0, x1, dimensions):
    '''
    Get the middle (M) coords of two coords.
    '''

    x0 = np.array(x0)
    x1 = np.array(x1)
    dimensions = np.array(dimensions)

    delta = x1 - x0
    
    delta2 = []
    for delt, box in zip(delta, dimensions):
        if delt < 0 and delt < 0.5 * box:
            delt = delt + box
        if delt > 0 and delt > 0.5 * box:
            delt = delt - box

        delta2.append(delt)


    delta2 = np.array(delta2)

    M_coords2 = 0.5 * delta2 + x0

    return M_coords2



def projectVectorToPlane(vector, plane_normal):
    '''
    Project a vector onto a plane, need normal of plane as input.
    '''

    plane_normal_magnitude = np.linalg.norm(plane_normal)
    
    project_vector = np.cross(vector, plane_normal)
    project_vector = np.divide(project_vector, plane_normal_magnitude)
    project_vector = np.cross(plane_normal, project_vector)
    projected_vector = np.divide(project_vector, plane_normal_magnitude)


    return projected_vector





def normaliseVector(x0, x1, dimensions):
    '''
    Get vector of two coordinates over PBCs and then normalise by the 
    magnitude of the vector
    '''

    x0 = np.array(x0)
    x1 = np.array(x1)
    dimensions = np.array(dimensions)
    delta = x1 - x0

    delta2 = []
    for delt, box in zip(delta, dimensions):
        if delt < 0 and delt < 0.5 * box:
            delt = delt + box
        if delt > 0 and delt > 0.5 * box:
            delt = delt - box

        delta2.append(delt)
        
    delta = np.array(delta2)

    delta_magnitude = (delta[0] ** 2 + delta[1] ** 2 + delta[2] ** 2) ** 0.5
    delta_norm = np.divide(delta, delta_magnitude)

    return delta_norm


def angleBetweenVectors(x0, x1):
    '''
    Get the angle between two vectors to determine how orthogonal they are.
    '''

    top = np.dot(x0, x1)

    axis1_magnitude = np.linalg.norm(x0) #magnitude/ distance
    axis2_magnitude = np.linalg.norm(x1) #magnitude/ distance
    bottom = np.multiply(axis1_magnitude, axis2_magnitude)
    top_bottom = np.divide(top, bottom)
    if top_bottom > 1:
        top_bottom = 1
    if top_bottom < -1:
        top_bottom = -1
    theta = np.arccos(top_bottom)
    theta = np.degrees(theta)

    return theta



def angle2(a, b, c, dimensions):
    '''
    Just used as a check. Doesn't work as periodic boundary conditions 
    are not considered properly.
    '''

    a = np.array(a) #j
    b = np.array(b) #i
    c = np.array(c) #k
    dimensions = np.array(dimensions)
    ba = a - b
    bc = c - b
    ac = c - a
    ba = np.where(ba > 0.5 * dimensions, ba - dimensions, ba)
    bc = np.where(bc > 0.5 * dimensions, bc - dimensions, bc)
    ac = np.where(ac > 0.5 * dimensions, ac - dimensions, ac)

    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))

    angle = np.arccos(cosine_angle)
    angle = np.degrees(angle)

    return angle



def angle(a, b, c, dimensions):

    a = np.array(a) #j
    b = np.array(b) #i
    c = np.array(c) #k
    dimensions = np.array(dimensions)
    ba = np.abs(a - b)
    bc = np.abs(c - b)
    ac = np.abs(c - a)
    ba = np.where(ba > 0.5 * dimensions, ba - dimensions, ba)
    bc = np.where(bc > 0.5 * dimensions, bc - dimensions, bc)
    ac = np.where(ac > 0.5 * dimensions, ac - dimensions, ac)
    dist_ba = np.sqrt((ba ** 2).sum(axis=-1))
    dist_bc = np.sqrt((bc ** 2).sum(axis=-1))
    dist_ac = np.sqrt((ac ** 2).sum(axis=-1))
    


    cosine_angle = ((dist_ac ** 2 - dist_bc ** 2 - dist_ba ** 2) /
                    (-2 * dist_bc * dist_ba))

    return cosine_angle





def com_vectors(coord_list, mass_list, dimensions):
    '''
    doesnt work
    '''
    cm = np.zeros(3)
    tot_mass = 0
    for coord, mass in zip(coord_list, mass_list):
        vec = vector(coord, [0, 0, 0], dimensions)
        cm += np.array(vec) * mass
        tot_mass += mass

    cm_weighted = np.array(cm) / float(tot_mass)

    return cm_weighted



def com(coord_list, mass_list):
    '''
    '''
    x_cm = 0
    y_cm = 0
    z_cm = 0
    tot_mass = 0
    for coord, mass in zip(coord_list, mass_list):
        x_cm += mass * coord[0]
        y_cm += mass * coord[1]
        z_cm += mass * coord[2]
        tot_mass += mass
    x_cm = float(x_cm) / float(tot_mass)
    y_cm = float(y_cm) / float(tot_mass)
    z_cm = float(z_cm) / float(tot_mass)

    cm = (x_cm, y_cm, z_cm)

    return cm




def MOI(cm, coord_list, mass_list):
    '''
    '''
    x_cm, y_cm, z_cm = cm[0], cm[1], cm[2]
    I_xx = 0
    I_xy = 0
    I_yx = 0

    I_yy = 0
    I_xz = 0
    I_zx = 0

    I_zz = 0
    I_yz = 0
    I_zy = 0

    for coord, mass in zip(coord_list, mass_list):
        I_xx += (abs(coord[1] - y_cm)**2 + abs(coord[2] - z_cm)**2) * mass
        I_xy += (coord[0] - x_cm) * (coord[1] - y_cm) * mass
        I_yx += (coord[0] - x_cm) * (coord[1] - y_cm) * mass

        I_yy += (abs(coord[0] - x_cm)**2 + abs(coord[2] - z_cm)**2) * mass
        I_xz += (coord[0] - x_cm) * (coord[2] - z_cm) * mass
        I_zx += (coord[0] - x_cm) * (coord[2] - z_cm) * mass

        I_zz += (abs(coord[0] - x_cm)**2 + abs(coord[1] - y_cm)**2) * mass
        I_yz += (coord[1] - y_cm) * (coord[2] - z_cm) * mass
        I_zy += (coord[1] - y_cm) * (coord[2] - z_cm) * mass

    I = np.array([[I_xx, -I_xy, -I_xz], [-I_yx, I_yy, -I_yz], 
        [-I_zx, -I_zy, I_zz]])
    return I



def UAforceTorque(bonded_atoms_list, F_axes, Fmoi_axes, MI_axes):
    '''
    Get UA force and torque from previously calculated Faxes and MIaxes
    '''

    '''
    Faxis_list = (Faxis1, Faxis2, Faxis3)
    f_list = (FO, FH1, FH2)
    MI_xyz = (axis1MI, axis2MI, axis3MI)
    water_torques = torque(c_list, m_list, Faxis_list, f_list, MI_xyz)
    '''

    c_list = []
    f_list = []
    forces_summed = np.zeros(3)
    m_list = []

    for atoms in bonded_atoms_list:
        c_list.append(atoms.coords)
        f_list.append(atoms.forces)
        forces_summed += atoms.forces
        m_list.append(atoms.mass)

    cm = com(c_list, m_list)
    UA_torque = torque_old(cm, c_list, Fmoi_axes, f_list, MI_axes)


    F1 = np.dot(forces_summed, F_axes[0])
    F2 = np.dot(forces_summed, F_axes[1])
    F3 = np.dot(forces_summed, F_axes[2])
    mass_sqrt = sum(m_list) ** 0.5
    UA_force = (float(F1)/float(mass_sqrt), 
            float(F2)/float(mass_sqrt), 
            float(F3)/float(mass_sqrt))


    return UA_force, UA_torque




def torque_old(cm, coord_list, axis_list, force_list, MI_coords):
    '''
    '''
    x_cm, y_cm, z_cm = cm[0], cm[1], cm[2]
    MI_coords = np.sqrt(MI_coords) #sqrt moi to weight torques

    O_torque = np.array([0, 0, 0])
    H1_torque = np.array([0, 0, 0])
    H2_torque = np.array([0, 0, 0])
    count_atom = 0
    for coord, force in zip(coord_list, force_list):
        count_atom +=1
        new_coords = []
        new_forces = []
        for axis in axis_list:
            new_c = axis[0]*(coord[0] - x_cm) + axis[1]*(coord[1] - y_cm) \
                    + axis[2]*(coord[2] - z_cm)
            new_f = axis[0]*force[0] + axis[1]*force[1] + axis[2]*force[2]
            new_coords.append(new_c)
            new_forces.append(new_f)

        torque_list = []
        torquex = float(new_coords[1]*new_forces[2] \
                - new_coords[2]*new_forces[1]) #/ float(1e10)
        torquey = float(new_coords[2]*new_forces[0] \
                - new_coords[0]*new_forces[2]) #/ float(1e10)
        torquez = float(new_coords[0]*new_forces[1] \
                - new_coords[1]*new_forces[0]) #/ float(1e10)

        if count_atom == 1:
            O_torque = (torquex, torquey, torquez)
            O_torque = np.divide(O_torque, MI_coords)
        elif count_atom == 2:
            H1_torque = (torquex, torquey, torquez)
            H1_torque = np.divide(H1_torque, MI_coords)
        elif count_atom == 3:
            H2_torque = (torquex, torquey, torquez)
            H2_torque = np.divide(H2_torque, MI_coords)
        else:
            ("torques outs of range")
            break

    W_torque = O_torque + H1_torque + H2_torque

    return (W_torque)



def principalAxesMOI(bonded_atoms_list):
    '''
    Calculate the principal axes of the MOI matrix, then calc forces
    '''

    coord_list = []
    mass_list = []
    forces_summed = np.zeros(3)

    for atom in bonded_atoms_list:
        coord_list.append(atom.coords)
        mass_list.append(atom.mass)
        forces_summed += atom.forces

    cm = com(coord_list, mass_list)
    moI = MOI(cm, coord_list, mass_list)


    eigenvalues, eigenvectors = LA.eig(moI) 
    #different values generated to Jon's code

    transposed = np.transpose(eigenvectors) #turn columns to rows

    min_eigenvalue = abs(eigenvalues[0])
    if eigenvalues[1] < min_eigenvalue:
        min_eigenvalue = eigenvalues[1]
    if eigenvalues[2] < min_eigenvalue:
        min_eigenvalue = eigenvalues[2]

    max_eigenvalue = abs(eigenvalues[0])
    if eigenvalues[1] > max_eigenvalue:
        max_eigenvalue = eigenvalues[1]
    if eigenvalues[2] > max_eigenvalue:
        max_eigenvalue = eigenvalues[2]

    #print (min_eigenvalue * const, max_eigenvalue * const) 
    #same as Jons

    FMIaxis1 = None
    FMIaxis2 = None
    FMIaxis3 = None

    axis1MI = None
    axis2MI = None
    axis3MI = None

    for i in range(0,3):
        if eigenvalues[i] == max_eigenvalue:
            FMIaxis1 = transposed[i]
            axis1MI = eigenvalues[i]
        elif eigenvalues[i] == min_eigenvalue:
            FMIaxis3 = transposed[i]
            axis3MI = eigenvalues[i]
        else:
            FMIaxis2 = transposed[i]
            axis2MI = eigenvalues[i]


    Fm_axes = [FMIaxis1, FMIaxis2, FMIaxis3]
    MI_axes = [axis1MI, axis2MI, axis3MI]

    return Fm_axes, MI_axes




def calcAngleWithNearestNonlike(neighbour, nearest, dimensions):
    '''
    Calc angle in plane and orthog to plane between any molecule
    and any other molecule.
    '''

    PAxes = np.array(neighbour.WMprincipalAxis[0])
    PAxis = np.array(neighbour.WMprincipalAxis[1])
    COM = np.array(neighbour.WMprincipalAxis[2])

    #plane_normal = vector(COM, PAxes[0], dimensions)
    #plane_normal2 = vector(COM, PAxes[1], dimensions)
    #OM_vector = vector(COM, PAxes[2], dimensions)

    plane_normal = PAxes[0]
    plane_normal2 = PAxes[1]
    OM_vector = PAxes[2]

    OSolute_vector = vector(nearest.coords, COM, dimensions)


    projected_OSolute = projectVectorToPlane(OSolute_vector, plane_normal)
    projected_orthog_OSolute = projectVectorToPlane(OSolute_vector, 
            plane_normal2)

    #print ('OM', OM_vector, O.atom_num)

    soluteOMangle = angleBetweenVectors(OM_vector, projected_OSolute)
    soluteOorthogAngle = angleBetweenVectors(OM_vector, 
            projected_orthog_OSolute)


    if np.dot(OM_vector, projected_OSolute) > 0 and \
            np.dot(plane_normal2, projected_OSolute) > 0:
        soluteOMangle = (90 - soluteOMangle) + 270

    if np.dot(-OM_vector, projected_OSolute) > 0 and \
            np.dot(plane_normal2, projected_OSolute) > 0:
        soluteOMangle = (180 - soluteOMangle) + 180


    if np.dot(-plane_normal, projected_orthog_OSolute) > 0 and \
            np.dot(-OM_vector, projected_orthog_OSolute) > 0:
        soluteOorthogAngle = (180 - soluteOorthogAngle) + 180

    if np.dot(-plane_normal, projected_orthog_OSolute) > 0 and \
            np.dot(OM_vector, projected_orthog_OSolute) > 0:
        soluteOorthogAngle = (90 - soluteOorthogAngle) + 270




    #cosAngle = angle(closest_H.coords, O.coords, solute.coords, dimensions)
    #arccosAngle = np.arccos(cosAngle)
    #angleDegrees = np.degrees(arccosAngle)
    #if math.isnan(angleDegrees) is False:
    #angleDegrees = int(round(angleDegrees, 0))

    if math.isnan(soluteOMangle) is False:
        soluteOMangle = int(round(soluteOMangle, 0))
    #else:
        #print ('Error1: NaN for solute resid: %s, '\
                #'water resid %s and angle %s.'\
                #% (nearest.resid, neighbour.resid, soluteOMangle))
        #continue

    if math.isnan(soluteOorthogAngle) is False:
        soluteOorthogAngle = int(round(soluteOorthogAngle, 0))
    #else:
        #print ('Error1: NaN for solute resid: %s, '\
                #'water resid %s and angle %s.'\
                #% (nearest.resid, neighbour.resid, soluteOorthogAngle))
        #continue


    return soluteOMangle, soluteOorthogAngle






def calcWaterSoluteAngle(solute, O, H1, H2, dimensions):
    '''
    Calc angle in plane and orthog to plane between any water
    and any solute atom.
    '''


    dist_solute_H1, dist_solute_H2 = None, None

    for atomDist in solute.nearest_all_atom_array:
        if atomDist[0] == H1.atom_num:
            dist_solute_H1 = atomDist[1]
        elif atomDist[0] == H2.atom_num:
            dist_solute_H2 = atomDist[1]
        else:
            continue

    if dist_solute_H1 == None:
        dist_solute_H1 = distance(solute.coords, H1.coords, dimensions)
    if dist_solute_H2 == None:
        dist_solute_H2 = distance(solute.coords, H2.coords, dimensions)
    closest_H = None
    if dist_solute_H1 < dist_solute_H2:
        closest_H = H1
    elif dist_solute_H1 > dist_solute_H2:
        closest_H = H2
    else:
        closest_H = H1


    H1O_vector = vector(O.coords, H1.coords, dimensions)
    H2O_vector = vector(O.coords, H2.coords, dimensions)
    plane_normal = np.cross(H1O_vector, H2O_vector)
    M_coords = getMcoords(H1.coords, H2.coords, dimensions)
    OM_vector = vector(M_coords, O.coords, dimensions)
    plane_normal2 = np.cross(OM_vector, plane_normal)
    OSolute_vector = vector(solute.coords, O.coords, dimensions)
    projected_OSolute = projectVectorToPlane(OSolute_vector, plane_normal)
    projected_orthog_OSolute = projectVectorToPlane(OSolute_vector, 
            plane_normal2)

    #print ('OM', OM_vector, O.atom_num)

    soluteOMangle = angleBetweenVectors(OM_vector, projected_OSolute)
    soluteOorthogAngle = angleBetweenVectors(OM_vector, 
            projected_orthog_OSolute)


    if np.dot(OM_vector, projected_OSolute) > 0 and \
            np.dot(plane_normal2, projected_OSolute) > 0:
        soluteOMangle = (90 - soluteOMangle) + 270
        ### only go up to 180 degrees rather than 360 as above
        #if soluteOMangle > 180:
            #soluteOMangle = 

    if np.dot(-OM_vector, projected_OSolute) > 0 and \
            np.dot(plane_normal2, projected_OSolute) > 0:
        soluteOMangle = (180 - soluteOMangle) + 180


    if np.dot(-plane_normal, projected_orthog_OSolute) > 0 and \
            np.dot(-OM_vector, projected_orthog_OSolute) > 0:
        soluteOorthogAngle = (180 - soluteOorthogAngle) + 180

    if np.dot(-plane_normal, projected_orthog_OSolute) > 0 and \
            np.dot(OM_vector, projected_orthog_OSolute) > 0:
        soluteOorthogAngle = (90 - soluteOorthogAngle) + 270




    cosAngle = angle(closest_H.coords, O.coords, solute.coords, dimensions)
    arccosAngle = np.arccos(cosAngle)
    angleDegrees = np.degrees(arccosAngle)
    if math.isnan(angleDegrees) is False:
        angleDegrees = int(round(angleDegrees, 0))
        soluteOMangle = int(round(soluteOMangle, 0))
        soluteOorthogAngle = int(round(soluteOorthogAngle, 0))
    else:
        logging.error('NaN for solute resid: %s, water resid %s and angle %s.'\
                % (solute.resid, O.resid, angleDegrees))


    return soluteOMangle, soluteOorthogAngle





"""
	A collection of utility functions or wrappers that 
	are used throughout the package.
"""
from __future__ import print_function

import struct
import numpy as nmp
import re
from datetime import datetime, timedelta

from CodeEntropy.FunctionCollection import UnitsAndConversions as CONST

def printflush(arg_string, end = '\n'):
	""" Modified print statement with flush = True in the end """
	print(arg_string,flush = True, end = end)
	return

#END

def printOut(arg_outFile, arg_string, end = "\n"):
	""" Modified print statement that will write a tring into a outfile """
	with open(arg_outFile, 'a') as fout:
		fout.write('{}{}'.format(arg_string, end))
	return

#END

def printList(arg_outFile,arg_list, arg_fmt, arg_maxRowLen, arg_descriptor = None):
	""" Writes the elements of a 1D list into the output file.
		At most `arg_maxRowLen` values are written per line in addition to the
		descriptor in each line. """
	listLength = len(arg_list)
	numWholeRows = int(nmp.floor(listLength/arg_maxRowLen))

	# if an empty list, print an empty line
	if listLength == 0:
		if arg_descriptor:
			printOut(arg_outFile,f"{arg_descriptor} ", '')
		printOut(arg_outFile,"")
		return

	startIdx = 1
	listIdx = 0
	for i in range(numWholeRows):
		if arg_descriptor:
			printOut(arg_outFile,f"{arg_descriptor} ", '')
		
		# printOut(arg_outFile,"{:>5d} {:>5d}".format(startIdx, startIdx + 5 - 1), ':')
		while listIdx < (startIdx + arg_maxRowLen - 1):
			printOut(arg_outFile,f"{arg_list[listIdx]:{arg_fmt}}",'')
			listIdx += 1
		printOut(arg_outFile,"")
		startIdx += arg_maxRowLen
	
	# the last row (if any)
	if (listLength % arg_maxRowLen != 0):
		if arg_descriptor:
			printOut(arg_outFile,f"{arg_descriptor} ", '')
		# printOut(arg_outFile,"{:>5d} {:>5d}".format(startIdx, listLength), ':')
		while listIdx < listLength:
			printOut(arg_outFile,f"{arg_list[listIdx]:{arg_fmt}}",'')
			listIdx += 1
		printOut(arg_outFile,"")

	return
#END

def hbar(arg_len = 50):
	printflush('-'*arg_len)
	return
#END

def print_time(arg_datetime):
	"""
	Prints date and time in YYYY-MM-DD HH:MM:SS format
	"""
	year = arg_datetime.year
	month = arg_datetime.month
	day = arg_datetime.day
	hour = arg_datetime.hour
	minute = arg_datetime.minute
	second = arg_datetime.second
	
	datestr = f"{year:>4d}-{month:>02d}-{day:>02d}  "
	timestr = f"{hour:>02d}:{minute:>02d}:{second:>02d}"
	printflush(datestr + timestr)
	return
#END

def print_duration(arg_t0, arg_t1):
	"""
	Prints the duration of time between t0 (first argument) and t1 (second argument)
	in appropriate human-readable format.
	"""
	dur = datetime(year = 1, month = 1, day = 1) + (arg_t1 - arg_t0)
	if dur.year > 1:
		timestr = f"{(dur.year - 1):>4d}yr {(dur.month - 1):>2d}mo {(dur.day - 1):>2d}dy {dur.hour:>2d}hr {dur.minute:>2d}min {dur.second:>2d}sec"
	elif dur.month > 1:
		timestr = f"{(dur.month - 1):>2d}mo {(dur.day - 1):>2d}dy {dur.hour:>2d}hr {dur.minute:>2d}min {dur.second:>2d}sec"
	elif dur.day > 1:
		timestr = f"{(dur.day - 1):>2d}dy {dur.hour:>2d}hr {dur.minute:>2d}min {dur.second:>2d}sec"
	elif dur.hour > 1:
		timestr = f"{dur.hour:>2d}hr {dur.minute:>2d}min {dur.second:>2d}sec"
	elif dur.minute > 1:
		timestr = f"{dur.minute:>2d}min {dur.second:>2d}sec"
	else:
		timestr = f"{dur.second:>2d}sec"

	printflush(timestr)
	return
#END


def decode(arg_fmt, arg_fileHandler):
	""" Using the struct module, read/decode a segment of binary data into a type indicated by the input format """
	sfmt = struct.calcsize(arg_fmt)
	return struct.unpack(arg_fmt, arg_fileHandler.readline(sfmt))
#END

def binary_to_dec_repr(arg_byteArray):
	""" For an input sequence of 0/1, return its equivalent decimal form in base 10 """
	decEquivalent = 0
	for idx, iPow in enumerate(range(len(arg_byteArray))):
		decEquivalent += (arg_byteArray[idx] * nmp.power(2,iPow))

	return decEquivalent
#END

def coalesce_numeric_array(arg_numArray):
	""" Take the elements in a given input array with integer elements and coalesce them to returna string whose characters
	are string cast of teh elements """
	charList = [str(char) for char in arg_numArray.astype(int)]
	return ''.join(charList)
#END


def diagonalize(arg_matrix):
	""" 
	Find the eigen vectors and values for the arg_matrix 
	using the linalg module of numpy.
	"""
	eigenValues, eigenVectors = nmp.linalg.eigh(arg_matrix)
	# a work around for complex values changing all very small value and negative to zero
	# eigenValues[eigenValues < 0 ] = 0
	return eigenValues, eigenVectors
#END 

def get_frequencies_from_nmd(arg_nmdfile, wnum2freq = True):
	"""
	From the NMD spectra file, extract frequency values and return a list.
	"""
	freqList = []
	with open(arg_nmdfile, "r") as fnmd:
		for line in fnmd:
			if re.match("^# \*\*\*", line):
				freq = float(line.split()[-1])
				if wnum2freq:
					freq *= CONST.C_LIGHT
				freqList.append(freq)
	return freqList
#END


import sys
import logging
import gc

from collections import Counter
from datetime import datetime

from CodeEntropy.poseidon.extractData.readFiles import populateTopology, getCoordsForces, getDistArray
# # Energy is not needed
# from CodeEntropy.poseidon.extractData.readFiles import populateEnergy, UAEnergyGroup
from CodeEntropy.poseidon.extractData.HBRAD import distCutoffNc, UALevelRAD, HBCalc
from CodeEntropy.poseidon.extractData.forceTorques import calculateFTMatrix
from CodeEntropy.poseidon.extractData.dihedrals import calculateDihedrals
from CodeEntropy.poseidon.extractData.nearestNonlike2 import getShellAssignment, moleculePositionRankingRAD 
from CodeEntropy.poseidon.extractData.outputFiles import moleculeObjectPopulation
from CodeEntropy.poseidon.extractData.mainClass import clearClass

from CodeEntropy.poseidon.analysis.populateClasses import classPopulation
from CodeEntropy.poseidon.analysis.EECalculation import processEE
from CodeEntropy.poseidon.analysis.helper import memoryInfo, weightingPopulation

from functools import partial
from multiprocessing import Pool

class Poseidon():
	"""
	Container to host data from MDAnalysis.Universe and run Poseidon Analysis
	"""

	def __init__(self ,container, start=0, end=-1, 
		step=1, pureAtomNum=1, cutShell=None, 
		excludedResnames=None,
		water='WAT', verbose=False):
		"""
		This is a initialization function to collect information from a MDanalysis universe into a data container for analysis using POSEIDON
		
		Parameters
		----------
		container : MDAnalyse.Universe
			A Universe object will all topology, dihedrals,coordinates and force information Check ``Example/create_new_universe.py`` on how to create a universe from existing data.
		start : int or None, Optional, default: 0
			Frame id to start analysis.
		end : int or None, Optional, default: -1
			Frame id to end analysis.
		step : int, Optional, default: 1
			Steps between frame.
		pureAtomNum : int, Optional, default: 1
			Reference molecule resid for pure liquid.
		cutShell : float or None, Optional, default: 1
			Explicit cut off shell. Default to None which uses the relative angular distance (RAD) algorithm. See Jonathan Higham and Richard H. Henchman , "Locally adaptive method to define coordination shell", J. Chem. Phys. 145, 084108 (2016)
		excludedResnames : list of str or None, Optional, default: None
			List of resnames to exclude from nearest non alike analysis.
		water : list of str, Optional, default: "WAT"
			Resname for water molecules.
		verbose : bool, Optional, default: False
			Print out progress of each step.
		
		"""
		startTime = datetime.now()
		print(startTime)
		verbosePrint = print if verbose else lambda *a, **k: None
		
		waterTuple = ('SOL', 'WAT', 'HOH', 'TIP3') #needed for pdb as top file 
		if water != 'WAT':
			waterTuple = (water)

		iterations = 0
		
		all_data = []

		populateTopology(container, all_data, waterTuple)
		verbosePrint('TOPOLOGY')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()


		resids = Counter([(i.resname) for i in all_data])
		verbosePrint(resids.keys())
		if len(resids.keys()) == 1:
			verbosePrint('Pure system with reference ID: %s' % (pureAtomNum))

		if excludedResnames != None:
			verbosePrint('EXCLUDED RESNAMES: %s' % (excludedResnames))
		

		dimensions = None
		allMoleculeList = []
		
		#fix end frame
		if end == -1:
			end = len(container.trajectory)

		for frame in range(int(start), int(end), int(step)):

			iterations += 1

			clearClass(all_data)

			print(f"frame = {frame}")

			all_data, dimensions = getCoordsForces(container, 
					all_data, dimensions, frame, startTime, 
					verbosePrint)
			# # Energy is not needed
			# populateEnergy(container, all_data, 
			#         dimensions, frame, startTime, verbosePrint)
			# UAEnergyGroup(all_data)


			calculateDihedrals(all_data, dimensions)
			verbosePrint('DIH')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush()

			#'''
			traj = container.trajectory[frame]
			neighbour_coords = None
			neighbour_coords = traj.positions

			max_cutoff = 10
			for x in range(0, len(all_data)):
				atom = all_data[x]
				#start nearest array from solutes
				if atom.resname not in waterTuple:
					getDistArray(atom, all_data, traj, max_cutoff,
							dimensions, neighbour_coords,
							startTime, verbosePrint)
					#find nearest array for solute neighbours that are solvent
					for nearDist in atom.nearest_all_atom_array[0:20]:
						neighbour = all_data[nearDist[0]]
						if neighbour.resname in waterTuple and \
								neighbour.nearest_all_atom_array == None:
							getDistArray(neighbour, all_data, traj, max_cutoff,
									dimensions, neighbour_coords,
									startTime, verbosePrint)
						#find solvent neighbours neighbours nearest array
						if neighbour.nearest_all_atom_array != None:
							for nearDist2 in neighbour.nearest_all_atom_array[0:20]:
								neighbour2 = all_data[nearDist2[0]]
								if neighbour2.resname in waterTuple and \
										neighbour2.nearest_all_atom_array == None:
									getDistArray(neighbour2, all_data, 
											traj, max_cutoff,
											dimensions, 
											neighbour_coords,
											startTime, verbosePrint)
								else:
									continue
						else:
							continue
				else:
					continue
			verbosePrint('NEAREST ARRAYS')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush()
			#'''



			if cutShell != None:
				#Used for fixed cut-off coordination shells
				try:
					cutoff_dist = float(cutShell)
					distCutoffNc(all_data, dimensions, cutoff_dist)
					# #don't know what this is it is not defined anywhere
					# NcPairs(all_data, dimensions)
					verbosePrint('distCutoffNc')
					verbosePrint(datetime.now() - startTime)
					sys.stdout.flush() 
				except ValueError:
					logging.error('Cutoff distance needs to be a float')


			if cutShell == None:
				UALevelRAD(all_data, dimensions)
				verbosePrint('RAD')
				verbosePrint(datetime.now() - startTime)


			#if inputType != 'pdb':
			HBCalc(all_data, waterTuple, dimensions)
			verbosePrint('HB')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush() 


			getShellAssignment(all_data, excludedResnames, dimensions, 
					startTime, verbosePrint)
			verbosePrint('PROX')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush()


			#if (force != None):
			calculateFTMatrix(all_data, dimensions)
			verbosePrint('FTMATRIX')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush()


			moleculePositionRankingRAD(all_data, waterTuple, dimensions)
			verbosePrint('ORIENTS')
			verbosePrint(datetime.now() - startTime)
			sys.stdout.flush()

			'''
			if iterations == 1:
				verbosePrint('Generating .pdb file for frame %s' % (frame+1))
				pdbGenerate(all_data, ('frame_%s' % (frame+1)), dimensions)
			'''



			allMoleculeList = moleculeObjectPopulation(all_data, 
					allMoleculeList, frame, dimensions)


			print(datetime.now() - startTime)
			sys.stdout.flush() #write out to output file for csf in 
					#real time rather than at end of job


		# #writing file here 
		# with gzip.GzipFile('moleculeListAll.obj', 'wb') as pickleFile:
		#     pickle.dump((allMoleculeList), pickleFile, protocol=2)
		#     pickleFile.close()

		print(datetime.now() - startTime)
			
		self.allMoleculeList = allMoleculeList

	def run_analysis(self, 
			temperature=298.0, 
			entropyEnergy=True, 
			level_list=['moleculeLevel'], 
			solvent=None, water='WAT',
			verbose=False, weighting=None, 
			forceUnits='kJ'):
		"""Perform analysis using poseidon

		Parameters
		-----------
		temperature : float, Optional, default: 298.0
			Temperature for system.
		entropyEnergy : Bool, Optional, default: True
			Run entropy and energy analysis.
		level_list : list of str, Optional, default: ['moleculeLevel']
			Choose and refine the level of analyiss: moleculeLevel, residLevel_resname, atomLevel, soluteContacts.
		solvent : str, Optional, default: None
			Resname for solvent.
		water : str or list of str, Optional, default: 'WAT'
			Resname for water.
		verbose : bool, Optional, default: False
			Print out progress of each step.
		weighting : str or None, Optional, default: None
			Get weighing for each frame if the simulation is biased.
		forceUnits : str, Optional, default: 'kJ'
			Units of forces, kJ or Kcal. 

		Returns
		---------
			Tuple of DataFrame, for each level analysed:
				solventData: dataframe that contains the result for solvent
				soluteData: dataframe that contains the result for solute
				contactMatrix: dataframe that contains the contact matrix
		"""

		verbosePrint = print if verbose else lambda *a, **k: None
		startTime = datetime.now()
		print(startTime)



		waterTuple = ['WAT', 'wat', 'SOL', 'H2O', 'h2o', 'WAT_O', 'TIP3']
		if water != 'WAT':
			waterTuple = [water]
		if solvent == None: ##when solvent is NOT water
			solvent = waterTuple




		print('\nsolvent: %s' % (solvent))
		print('\nwater: %s' % (waterTuple))
		print('\n1. Populate Dictionaries\n')

		count = 1
		totAtoms = 0
		list_len = None
		atom_nums = []
		atom_count = 0
		objectIteration = 0

		EEclass = None 
		EEclass_residLevel_resname = None
		EEclass_soluteContacts = None
		EEclass_atomLevel = None


		class_str_list = ['EEclass', 'EEclass_residLevel_resname', 
				'EEclass_soluteContacts', 'EEclass_atomLevel', 
				'totAtoms', 'atom_nums', 'atom_count', 'objectIteration']

		memoryInfo(verbosePrint)
		print(datetime.now() - startTime)
		sys.stdout.flush()
		if totAtoms != 0:
			count = 2

		weighting_info = None
		atomList = self.allMoleculeList
		list_len = len(atomList[0])
		totAtoms += len(atomList)
		if count == 1:
			for i in atomList:
				atom_num = i[2]
				if atom_num not in atom_nums:
					atom_nums.append(atom_num)
				else:
					break
			if weighting != None:
				weighting_info = weightingPopulation(weighting)

		EEclass, EEclass_residLevel_resname, \
				EEclass_soluteContacts, EEclass_atomLevel, \
				atom_count = \
					classPopulation(
				atomList, entropyEnergy, 
				level_list, count, 
				atom_count, len(atom_nums), 
				waterTuple, temperature, 
				solvent, EEclass, 
				EEclass_residLevel_resname, EEclass_soluteContacts, 
				EEclass_atomLevel, 
				weighting_info, verbosePrint)
		count += 1 #needed for class initiation
		gc.enable()
		memoryInfo(verbosePrint)
		print(datetime.now() - startTime)
		sys.stdout.flush()


		print(datetime.now() - startTime)
		memoryInfo(verbosePrint)
		sys.stdout.flush()

		
		try:
			totFrames = float(totAtoms) / float(len(atom_nums))
			print('\nTotal number of frames: %s' % (totFrames))
			print('Number of atoms in each frame: %s' % (len(atom_nums)))
			print('Number of variables in each list: %s' % (list_len))
			num_frames = totFrames
		except ZeroDivisionError:
			logging.error('No frames to analyse, please chose correct '\
					'path to .obj files')
			num_frames = None
			sys.exit()

		###once all the classes have been populated, calculate properties
		#and output to files
		
		print('\n2. Process Dictionaries')
		result_list = dict()
		print(level_list)
		for level in level_list:

			print('---level:', level)
			EEclass2, DSclass2 = None, None

			if level in [None, 'moleculeLevel']:
				EEclass2 = EEclass
			if level in ['residLevel_resname']:
				EEclass2 = EEclass_residLevel_resname
			if level in ['soluteContacts']:
				EEclass2 = EEclass_soluteContacts
			if level in ['atomLevel']:
				EEclass2 = EEclass_atomLevel

			
			if entropyEnergy:
				name = 'EE'
				result = processEE(num_frames, totFrames, EEclass2, 
						solvent, waterTuple, 
						temperature, level, name, forceUnits, verbosePrint)
				result_list[level] = result
		# #'''
		# ##Save each class as an object so that we can continue populating
		# #in stages.
		# if len(paths) != 0:
		#     objectIteration += 1
		#     if pathClasses:
		#         print('\n3. Save Dictionaries')
		#         for Aclass in class_str_list:
		#             with gzip.GzipFile('%s.obj' % (Aclass), 'wb') as pickleFile:
		#                 pickle.dump((locals()[Aclass]), pickleFile, protocol=2)
		#                 pickleFile.close()
		#                 #locals()[Aclass] = None
		#         print('Number of objectIteration cycles saved: %s' % 
		#                 (objectIteration))
		#         print('Total atoms processed: %s' % (totAtoms))

		# #'''


		sys.stdout.flush()
		print('\n')
		print(datetime.now() - startTime)
		return result_list

class Poseidon_mp(Poseidon):
	def __init__(self ,container, start=0, end=-1, 
		step=1, pureAtomNum=1, cutShell=None, 
		excludedResnames=None,
		water='WAT', verbose=False, thread=4):
		"""This is a initialization function to collect information from a MDanalysis universe into a data container for analysis using POSEIDON
		Pending Work
			- rewrite this into a class initalization for easier understanding
			- rewrite the energy part to be compulsory

		Args:
			container (MDAnalysis.Universe): A MDAnalysis object with coordinates, forces and energy (loaded to velocity field with the value [Potential Energy(index 0), Kinetic Energy(1) and dummy(2)] respectively)
			start (int): Starting Frame ID. Defaults to 0.
			end (int): Ending Frame ID, this frame is not included. Defaults to -1.
			step (int, optional): Steps between frame. Defaults to 1.
			pureAtomNum (int, optional): Reference molecule resid for pure liquid. Defaults to 1.
			cutShell (float, optional): Explicit cut off shell (might be buggy since part of it is not defined). Default to None which uses the relative angular distance (RAD) algorithm. See Jonathan Higham and Richard H. Henchman , "Locally adaptive method to define coordination shell", J. Chem. Phys. 145, 084108 (2016)
			excludedResnames (list, optional): List of resnames to exclude from nearest non alike analysis. Defaults to None.
			water (str, optional): Resname for water molecules. Defaults to 'WAT'.
			verbose (bool, optional): print out progress of each analysis step. Defaults to False.
		"""
		startTime = datetime.now()
		print(startTime)
		verbosePrint = print if verbose else lambda *a, **k: None
		
		waterTuple = ('SOL', 'WAT', 'HOH', 'TIP3') #needed for pdb as top file 
		if water != 'WAT':
			waterTuple = (water)

		iterations = 0
		
		all_data = []

		populateTopology(container, all_data, waterTuple)
		verbosePrint('TOPOLOGY')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()


		resids = Counter([(i.resname) for i in all_data])
		verbosePrint(resids.keys())
		if len(resids.keys()) == 1:
			verbosePrint('Pure system with reference ID: %s' % (pureAtomNum))

		if excludedResnames != None:
			verbosePrint('EXCLUDED RESNAMES: %s' % (excludedResnames))
		

		dimensions = None
		allMoleculeList = []
		
		#fix end frame
		if end == -1:
			end = len(container.trajectory)
		sys.setrecursionlimit(3000000)
		p = Pool(processes=thread)
		fi_partial = partial(self.frame_iteration, container, all_data, dimensions, startTime, verbosePrint, waterTuple, cutShell, excludedResnames)
		data = p.map(fi_partial, [frame for frame in range(int(start), int(end), int(step))])
		# data = []
		# for frame in range(int(start), int(end), int(step)):
		#     data.append(fi_partial(frame))
		for item in data:
			allMoleculeList = moleculeObjectPopulation(item[0], 
					allMoleculeList, item[1], item[2])
		# #writing file here 
		# with gzip.GzipFile('moleculeListAll.obj', 'wb') as pickleFile:
		#     pickle.dump((allMoleculeList), pickleFile, protocol=2)
		#     pickleFile.close()

		print(datetime.now() - startTime)
			
		self.allMoleculeList = allMoleculeList
	
	def frame_iteration(self, container, all_data, dimensions, startTime, verbosePrint, waterTuple, cutShell, excludedResnames, frame):
		clearClass(all_data)
		print(f"frame = {frame}")

		all_data, dimensions = getCoordsForces(container, 
				all_data, dimensions, frame, startTime, 
				verbosePrint)
		# # Energy is not needed
		# populateEnergy(container, all_data, 
		#         dimensions, frame, startTime, verbosePrint)
		# UAEnergyGroup(all_data)


		calculateDihedrals(all_data, dimensions)
		verbosePrint('DIH')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()

		#'''
		traj = container.trajectory[frame]
		neighbour_coords = None
		neighbour_coords = traj.positions

		max_cutoff = 10
		for x in range(0, len(all_data)):
			atom = all_data[x]
			#start nearest array from solutes
			if atom.resname not in waterTuple:
				getDistArray(atom, all_data, traj, max_cutoff,
						dimensions, neighbour_coords,
						startTime, verbosePrint)
				#find nearest array for solute neighbours that are solvent
				for nearDist in atom.nearest_all_atom_array[0:20]:
					neighbour = all_data[nearDist[0]]
					if neighbour.resname in waterTuple and \
							neighbour.nearest_all_atom_array == None:
						getDistArray(neighbour, all_data, traj, max_cutoff,
								dimensions, neighbour_coords,
								startTime, verbosePrint)
					#find solvent neighbours neighbours nearest array
					if neighbour.nearest_all_atom_array != None:
						for nearDist2 in neighbour.nearest_all_atom_array[0:20]:
							neighbour2 = all_data[nearDist2[0]]
							if neighbour2.resname in waterTuple and \
									neighbour2.nearest_all_atom_array == None:
								getDistArray(neighbour2, all_data, 
										traj, max_cutoff,
										dimensions, 
										neighbour_coords,
										startTime, verbosePrint)
							else:
								continue
					else:
						continue
			else:
				continue
		verbosePrint('NEAREST ARRAYS')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()
		#'''



		if cutShell != None:
			#Used for fixed cut-off coordination shells
			try:
				cutoff_dist = float(cutShell)
				distCutoffNc(all_data, dimensions, cutoff_dist)
				# #don't know what this is it is not defined anywhere
				# NcPairs(all_data, dimensions)
				verbosePrint('distCutoffNc')
				verbosePrint(datetime.now() - startTime)
				sys.stdout.flush() 
			except ValueError:
				logging.error('Cutoff distance needs to be a float')


		if cutShell == None:
			UALevelRAD(all_data, dimensions)
			verbosePrint('RAD')
			verbosePrint(datetime.now() - startTime)


		#if inputType != 'pdb':
		HBCalc(all_data, waterTuple, dimensions)
		verbosePrint('HB')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush() 


		getShellAssignment(all_data, excludedResnames, dimensions, 
				startTime, verbosePrint)
		verbosePrint('PROX')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()


		#if (force != None):
		calculateFTMatrix(all_data, dimensions)
		verbosePrint('FTMATRIX')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()


		moleculePositionRankingRAD(all_data, waterTuple, dimensions)
		verbosePrint('ORIENTS')
		verbosePrint(datetime.now() - startTime)
		sys.stdout.flush()

		'''
		if iterations == 1:
			verbosePrint('Generating .pdb file for frame %s' % (frame+1))
			pdbGenerate(all_data, ('frame_%s' % (frame+1)), dimensions)
		'''

		return (all_data, frame, dimensions)



#!/usr/bin/env python

import sys
import math
import numpy as np


from CodeEntropy.poseidon.extractData.generalFunctions import *


def calculateDihedrals(all_data, dimensions):
    '''
    In-place modification of all_data to calculate dihedrals
    For each dihedral in system, find what conformation it is in
    Dihedrals:


               X4
               /
              /
       X2---X3
      /
     /
    X1


    b1 = vector(X1, X2)
    b2 = vector(X2, X3)
    b3 = vector(X3, X4)

    source = https://math.stackexchange.com/questions/47059/
        how-do-i-calculate-a-dihedral-angle-given-cartesian-coordinates 

    Args:
        all_data (list of poseidon.atom_info): Container for all atom info
        dimensions (numpy array of size (1,3)): dimension of box of current frame
    '''

    rad2deg = float(180)/float(np.pi)

    dihedrals = []
    all_atom_nums_list = []

    max_atom_num = max(all_data[x].atom_num
            for x in range(0, len(all_data)))


    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1 and len(atom.dihedral_list) != 0 and \
                atom.atom_num not in all_atom_nums_list:

            Mdihedrals = [] #unique dihedrals for each atom in molecule
            lowest_atom_num = max_atom_num + 1
            for molecule_atom in atom.molecule_atomNums:
                Matom = all_data[molecule_atom]
                all_atom_nums_list.append(Matom.atom_num)
                #get lowest atom num for renumbering
                for dih in Matom.dihedral_list:
                    for d in dih:
                        if all_data[d].resid == atom.resid:
                            if d < lowest_atom_num:
                                lowest_atom_num = d
                            else:
                                continue

            for molecule_atom in atom.molecule_atomNums:
                Matom = all_data[molecule_atom]
                for dih in Matom.dihedral_list:

                    inside_resid = True
                    for d in dih:
                        if all_data[d].resid != atom.resid:
                            inside_resid = False

                    if dih not in dihedrals and inside_resid == True: 
                        #only analyse dihedral once
                        dihedrals.append(dih)

                        d_atoms = [] 
                        atom_nums = [] #1,2,3,4
                        for d in dih: #calc dih angles for each dih
                            atom2 = all_data[d]
                            d_atoms.append(atom2)
                            atom_nums.append(atom2.atom_num)

                        
                        #re-number atom_nums
                        renumbered_atom_nums = []
                        for num in atom_nums:
                            renumbered = (num - lowest_atom_num) + 1
                            renumbered_atom_nums.append(renumbered)

                        ##get vectors
                        b1_vector = vector(d_atoms[0].coords, 
                                d_atoms[1].coords, dimensions)
                        b1_dist = distance(d_atoms[0].coords, 
                                d_atoms[1].coords, dimensions)
                        b1_norm = np.divide(b1_vector, b1_dist)

                        b2_vector = vector(d_atoms[1].coords, 
                                d_atoms[2].coords, dimensions)
                        b2_dist = distance(d_atoms[1].coords, 
                                d_atoms[2].coords, dimensions)
                        b2_norm = np.divide(b2_vector, b2_dist)

                        b3_vector = vector(d_atoms[2].coords, 
                                d_atoms[3].coords, dimensions)
                        b3_dist = distance(d_atoms[2].coords, 
                                d_atoms[3].coords, dimensions)
                        b3_norm = np.divide(b3_vector, b3_dist)

                        ## get normals
                        n1 = np.cross(b1_norm, b2_norm)
                        n2 = np.cross(b2_norm, b3_norm)
                        m = np.cross(n1, b2_norm)

                        #get dot products
                        x = np.dot(n1, n2)
                        y = np.dot(m, n2)

                        phi = rad2deg * np.arctan2(y, x)
                        dih_type = None

                        ## only needed for fixed-dihedrals
                        if phi >= 120 or phi < -120:
                            dih_type = [0, 'trans']

                        if phi >= 0 and phi < 120:
                            dih_type = [1, 'g-']

                        if phi >= -120 and phi < 0:
                            dih_type = [2, 'g+']


                        Mdihedrals.append([renumbered_atom_nums, 
                            int(round(phi, 0)), 
                            dih_type])

                    else:
                        continue

            atom.dihedral_phi_type = Mdihedrals

        else:
            continue




import xdrlib
import numpy as nmp
import sys

def print_FTC_matrix_bin( arg_filename ):
	""" This function writes the FTC matrix (FF + FT + TF + TT) for a given instance of BeadCollection in binary format.
	The sequence of the data is:
	N (# Beads), M x M (length of FTC matrix), MFF x MFF (length of FF matrix), MFT_row x MFT_col (length of FT matrix),
	MTF_row x MTF_col (length of TF matrix), MTT x MTT (length of TT matrix),
	FF matrix as 1D array, FT matrix as 1D array, TF matrix as 1D array, TT matrix as 1D array. Dimensions are unsigned int
	and arrays are in floats """

	with open(arg_filename, 'rb') as fBin:

		buff = fBin.read()
		xup = xdrlib.Unpacker(buff)

		# number of beads
		numBeads = xup.unpack_uint()
		print('Number of beads = {}'.format( numBeads )) 

		# dim of FTC matrix
		shapeFTC = (xup.unpack_uint(), xup.unpack_uint())
		print('FTC matrix size = {} x {}'.format( *shapeFTC ))
		
		# dim of FF, FT, TF, TT
		matrixShapeDict = dict()
		for iMat in ["FF", "FT", "TF", "TT"]:
			matrixShapeDict[iMat] = (xup.unpack_uint(), xup.unpack_uint())
			print('{} matrix size = {} x {}'.format( iMat, *matrixShapeDict[iMat] ))

		# reshape the flattened version of  the submatrices into appropriate sizes
		matrixDict = dict()
		for iMat in ["FF", "FT", "TF", "TT"]:
			matrixSize = nmp.product( matrixShapeDict[iMat] )
			flatArr = [ xup.unpack_float() for i in range(matrixSize) ]
			matrixDict[iMat] = nmp.reshape( flatArr,  matrixShapeDict[iMat] )
			print('Trace of {} : {:>10.4}'.format( iMat, nmp.trace(matrixDict[iMat]) ))
		# place them in correct order
		matrixFTC = nmp.block( [     [matrixDict["FF"]  ,  matrixDict["FT"]],
		                             [matrixDict["TF"]  ,  matrixDict["TT"]]  ] )

		# assert that this is a square matrix
		assert(nmp.shape(matrixFTC) == shapeFTC)

		# print
		for iRow in range(shapeFTC[0]):
			for iCol in range(shapeFTC[1]):
				print('{:>10.4}'.format(matrixFTC[iRow, iCol]), end = '')
			print('')




	return
#END

if __name__ == "__main__":
	
	if len(sys.argv) != 2:
		print('Input a bindary FTC matrix file to print its contents out in the terminal')
		sys.exit()

	print_FTC_matrix_bin(sys.argv[1])
